#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Description: This example implements a running light. It shows various ways to access data
# of the process image with python

# Hardware setup: DIO module left of the RevPi module. Outputs have to start at byte 70.

import time    # used for the delays of the demo
import struct  # used for the processing of byte-strings espectially with Python 2.7
import fcntl   # used for the byte access of the process image within IOCTL

a=1 #  Integer which is shifted within 14 positions in order to generate a running light at the DIO module

# first the driver has to be opened by the "open" statement:
f = open("/dev/piControl0","wb+",0)


# now the endless loop of the demo starts ...
while 1:
    
	# Phase 1: Running light from Output_Pin_1 to Output_Pin_14 at the DIO
	# This phase shows the access which is possible with Python 2.7 and Python 3.
	# The struct library is used to generate a byte-string which is used by the 
	# write-function
    for i in range(0,14):
        f.seek(70)   # here the offset within the process image is set: Outputs start at byte 70 in the configuration
        x = struct.pack('<H',(a<<i)) 
        f.write(x)    # here 2 bytes are written into the process image, because with 'H' the pack function packs 2 bytes into x
        time.sleep(0.1)        

	# Phase 2: Running light from Output_Pin_14 through Output_Pin_1 at the DIO
	# This phase shows the access which is possible with Python 2.7 and Python 3.
	# Instead of struct here we are using a bytearray in order to generate a byte-string which is used by the write-function.
    for i in range(13,-1,-1):
        f.seek(70)
        x1 = 255 & (a<<i)       # low byte
        x2 = 255 & ((a<<i)>>8)  # high byte
        x =bytearray([x1,x2])
        f.write(x)
        time.sleep(0.1)

	# Phase 3: Running light from outside to the middle
	# This phase shows the access which is possible just with Python 3.
	# Starting with Python 3 the Integers provide the method "to_bytes"
	# which can be used to create a byte-string easy...
    for i in range(0,7):
        f.seek(70)
        x=((a<<i)|(a<<(13-i))).to_bytes(2, 'little') # wir brauchen little Endian für das Prozessabbild
        f.write(x)
        time.sleep(0.1)

	# Phase 4: Running light from the middle to outside
	# This phase shows the access which is possible just with Python 3.
	# Starting with Python 3 the function "bytes" may create a byte array 
	# from integers
    for i in range(6,-1,-1):
        f.seek(70)
        x1 = 255 & ((a<<i)|(a<<(13-i)))
        x2 = 255 & (((a<<i)|(a<<(13-i)))>>8)
        x = bytes([x1,x2])
        f.write(x)
        time.sleep(0.1)
		
	# Phase 5: the 14 inputs are mirrored to the outputs
	# we are reading the inputs now
    f.seek(0)      			# Input start at offset 0 in this configuration
    x = f.read(2)  			# 2 bytes are read
    #i = x[0] + 256*x[1]  		# This simple convertion of Byte-string into integers is possible only with Python 3!
    s = struct.unpack('<H',x) 	# this is the way which is possible with Python 2.7
    i = s[0]   				# because is is a list we have to restrict the access to the first element
    f.seek(70) 				# now we write the currently read data... Outputs start at offset 70 in this configuration
    #f.write(i.to_bytes(2,'little'))   # to_byte is just available in Python 3
    f.write(struct.pack('<H',i))     	# with Python 2.7 we need the struct-Library
    time.sleep(3)

	# Phase 6: We determine the offset via IOCTL and the bit of a symbolic name (which is created by PiCtory),
	# in order to read and write a sindle bit_
	# List of the available IOCTL-Functions (see C-Sourcecode of "piTest"):
 
	#define  KB_CMD1                    _IO(KB_IOC_MAGIC, 10 )  // for test only
	#define  KB_CMD2                    _IO(KB_IOC_MAGIC, 11 )  // for test only
	#define  KB_RESET                   _IO(KB_IOC_MAGIC, 12 )  // reset the piControl driver including the config file
	#define  KB_GET_DEVICE_INFO_LIST    _IO(KB_IOC_MAGIC, 13 )  // get the device info of all detected devices
	#define  KB_GET_DEVICE_INFO         _IO(KB_IOC_MAGIC, 14 )  // get the device info of one device
	#define  KB_GET_VALUE               _IO(KB_IOC_MAGIC, 15 )  // get the value of one bit in the process image
	#define  KB_SET_VALUE               _IO(KB_IOC_MAGIC, 16 )  // set the value of one bit in the process image
	#define  KB_FIND_VARIABLE           _IO(KB_IOC_MAGIC, 17 )  // find a varible defined in piCtory
	#define  KB_SET_EXPORTED_OUTPUTS    _IO(KB_IOC_MAGIC, 18 )  // copy the exported outputs from a application process image to the real process image
   
	# first function: We are searching the offset and the byte of "Input_Pin_4" using KB_FIND_VARIABLE
    prm = (b'K'[0]<<8) + 17   # the IOCTL-parameter is calculated using the ASCIIC 'K' which is shifted by 8 bits plus the ID of the desired function (=17)
    name = struct.pack('37s',b"I_1") # The arcument for the function 17 is a packed byte array with contains 37 bytes. The first 32 bytes consist of the symbolic name
    ret = fcntl.ioctl(f,prm, name)			  # The result values are written into a byte-array
    
	# in wich the first 32 byte are the symbolic name then 2 byts for the offset and then 1 byte for the bitposition and then 2 byte for the length.
	# In C it looks like that
	
	#  typedef struct
	#{
	#   char        strVarName[32]         // Variable name
	#    uint16_t    i16uAddress           // Address of the byte in the process image
	#    uint8_t     i8uBit                // 0-7 bit position, >= 8 whole byte
	#    uint16_t    i16uLength            // length of the variable in bits. Possible values are 1, 8, 16 and 32
	#} SPIVariable

    offset = struct.unpack_from('>H',ret,32)[0]   # this is also possible in Python 2.7
    bit = struct.unpack_from('B',ret,34)[0]
    length = struct.unpack_from('H',ret,35)[0]

	# now with the function 15 a single bit ("Input_Pin_4") is read out of the inputs...
    prm = (b'K'[0]<<8) + 15
	# The argument of the function is a bytearray with this stuct:
	
	#typedef struct
	#{
	#    uint16_t    i16uAddress            // Address of the byte in the process image
	#    uint8_t     i8uBit                 // 0-7 bit position, >= 8 whole byte
	#    uint8_t     i8uValue               // Value: 0/1 for bit access, whole byte otherwise
	#} SPIValue
    
    value = bytearray([0,0,0,0])
    struct.pack_into('>H',value,0,offset)    # this is also possible in Python 2.7
    struct.pack_into('B',value,2,bit)
    fcntl.ioctl(f,prm,value)
    bitval = value[3]
	
	# once more the function 17. Now we want to know, where "Output_Pin_4" is in the process image...
    prm = (b'K'[0]<<8) + 17
    name = struct.pack('37s',b"O_4")
    ret = fcntl.ioctl(f,prm, name)    
    offset = struct.unpack_from('>H',ret,32)[0]
    bit = struct.unpack_from('B',ret,34)[0]
    length = struct.unpack_from('H',ret,35)[0]
	
	# now with the function 16 a single bit ("Output_Pin_4") is written to the outputs...
    prm = (b'K'[0]<<8) + 16
    value = bytearray([0,0,0,0])
    struct.pack_into('>H',value,0,offset)
    struct.pack_into('B',value,2,bit)
    struct.pack_into('B',value,3,bitval)
    fcntl.ioctl(f,prm,value)
    
    time.sleep(3)
    